@@ -1428,7 +1428,7 @@ bool qpwgraph_canvas::saveState (void) const
 				m_settings->setValue('/' + node_key, node->pos());
 				m_settings->endGroup();
 				m_settings->beginGroup(NodeAliasesGroup);
-				if (node->nodeName() != node->nodeTitle())
+				if (node->nodeNameLabel() != node->nodeTitle())
 					m_settings->setValue('/' + node_key, node->nodeTitle());
 				else
 					m_settings->remove('/' + node_key);
@@ -1442,7 +1442,7 @@ bool qpwgraph_canvas::saveState (void) const
 			if (port) {
 				const QString& port_key = portKey(port);
 				m_settings->beginGroup(PortAliasesGroup);
-				if (port && port->portName() != port->portTitle())
+				if (port && port->portNameLabel() != port->portTitle())
 					m_settings->setValue('/' + port_key, port->portTitle());
 				else
 					m_settings->remove('/' + port_key);
