@@ -1,7 +1,7 @@
 // qpwgraph_pipewire.h
 //
 /****************************************************************************
-   Copyright (C) 2021-2022, rncbc aka Rui Nuno Capela. All rights reserved.
+   Copyright (C) 2021-2023, rncbc aka Rui Nuno Capela. All rights reserved.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
@@ -81,6 +81,7 @@ public:
 	// PipeWire client data struct access.
 	//
 	struct Data;
+	struct Proxy;
 	struct Object;
 	struct Node;
 	struct Port;
@@ -94,13 +95,16 @@ public:
 	void removeObject(uint id);
 	void clearObjects();
 
+	void addObjectEx(uint id, Object *object);
+	void removeObjectEx(uint id);
+
 	// Node methods....
 	Node *findNode(uint node_id) const;
 	Node *createNode(
 		uint node_id,
 		const QString& node_name,
 		qpwgraph_item::Mode node_mode,
-		uint node_types);
+		uint node_type);
 	void destroyNode(Node *node);
 
 	// Port methods....
@@ -144,7 +148,8 @@ private:
 	QList<Object *> m_objects;
 
 	// Callback sanity mutex.
-	static QMutex g_mutex;
+	QMutex m_mutex1;
+	QMutex m_mutex2;
 };
 
 
